// MyExt.cpp

#include "dbgexts.h"
#include "outputcallbacks.h"

/***************************************************************************************************************************************************************/
/*************************************************************** O U T C T L   E X A M P L E S *****************************************************************/
/***************************************************************************************************************************************************************/

HRESULT CALLBACK 
outctlpassed(PDEBUG_CLIENT pDebugClient, PCSTR args)
{
	UNREFERENCED_PARAMETER(args);

	IDebugControl* pDebugControl;
	if (SUCCEEDED(pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
	{
		pDebugControl->ControlledOutput(DEBUG_OUTCTL_THIS_CLIENT,       DEBUG_OUTPUT_NORMAL, "DEBUG_OUTCTL_THIS_CLIENT\n");
		pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_CLIENTS,       DEBUG_OUTPUT_NORMAL, "DEBUG_OUTCTL_ALL_CLIENTS\n");
		pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_OTHER_CLIENTS, DEBUG_OUTPUT_NORMAL, "DEBUG_OUTCTL_ALL_OTHER_CLIENTS\n");
		pDebugControl->ControlledOutput(DEBUG_OUTCTL_IGNORE,            DEBUG_OUTPUT_NORMAL, "DEBUG_OUTCTL_IGNORE\n");
		pDebugControl->ControlledOutput(DEBUG_OUTCTL_LOG_ONLY,          DEBUG_OUTPUT_NORMAL, "DEBUG_OUTCTL_LOG_ONLY\n");

		pDebugControl->Release();
	}
	return S_OK;
}

HRESULT CALLBACK 
outctlcreate(PDEBUG_CLIENT pDebugClient, PCSTR args)
{
	UNREFERENCED_PARAMETER(args);

	IDebugClient* pDebugCreated;
	if (SUCCEEDED(pDebugClient->CreateClient(&pDebugCreated)))
	{
		IDebugControl* pDebugControl;
		if (SUCCEEDED(pDebugCreated->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
		{
			pDebugControl->ControlledOutput(DEBUG_OUTCTL_THIS_CLIENT,       DEBUG_OUTPUT_NORMAL, "DEBUG_OUTCTL_THIS_CLIENT\n");
			pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_CLIENTS,       DEBUG_OUTPUT_NORMAL, "DEBUG_OUTCTL_ALL_CLIENTS\n");
			pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_OTHER_CLIENTS, DEBUG_OUTPUT_NORMAL, "DEBUG_OUTCTL_ALL_OTHER_CLIENTS\n");
			pDebugControl->ControlledOutput(DEBUG_OUTCTL_IGNORE,            DEBUG_OUTPUT_NORMAL, "DEBUG_OUTCTL_IGNORE\n");
			pDebugControl->ControlledOutput(DEBUG_OUTCTL_LOG_ONLY,          DEBUG_OUTPUT_NORMAL, "DEBUG_OUTCTL_LOG_ONLY\n");

			pDebugControl->Release();
		}
		pDebugCreated->Release();
	}
	return S_OK;
}

HRESULT CALLBACK 
outctlthis(PDEBUG_CLIENT pDebugClient, PCSTR args)
{
	UNREFERENCED_PARAMETER(args);

	IDebugClient* pDebugCreated;
	if (SUCCEEDED(pDebugClient->CreateClient(&pDebugCreated)))
	{
		IDebugControl* pDebugControl;
		if (SUCCEEDED(pDebugCreated->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
		{
			pDebugControl->ControlledOutput(DEBUG_OUTCTL_THIS_CLIENT | DEBUG_OUTCTL_NOT_LOGGED, DEBUG_OUTPUT_NORMAL, "DEBUG_OUTCTL_THIS_CLIENT | DEBUG_OUTCTL_NOT_LOGGED\n");
			pDebugControl->Release();
		}
		pDebugCreated->Release();
	}
	return S_OK;
}

HRESULT CALLBACK 
outctlexecute(PDEBUG_CLIENT pDebugClient, PCSTR args)
{
	IDebugClient* pDebugCreated;
	if (SUCCEEDED(pDebugClient->CreateClient(&pDebugCreated)))
	{
		IDebugControl* pDebugControl;
		if (SUCCEEDED(pDebugCreated->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
		{
			pDebugControl->Execute(DEBUG_OUTCTL_THIS_CLIENT | DEBUG_OUTCTL_NOT_LOGGED, args, DEBUG_EXECUTE_NOT_LOGGED | DEBUG_EXECUTE_NO_REPEAT);
			pDebugControl->Release();
		}
		pDebugCreated->Release();
	}
	return S_OK;
}

/***************************************************************************************************************************************************************/
/*********************************************************** C A L L B A C K   E X A M P L E S *****************************************************************/
/***************************************************************************************************************************************************************/

HRESULT CALLBACK 
callbackecho(PDEBUG_CLIENT pDebugClient, PCSTR args)
{
	IDebugClient* pDebugCreated;
	if (SUCCEEDED(pDebugClient->CreateClient(&pDebugCreated)))
	{
		IDebugControl* pDebugControl;
		if (SUCCEEDED(pDebugCreated->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
		{
			// Create our IDebugOutputCallbacks based object
			#pragma warning( disable : 28197 ) // PreFAST sees the 'new' as a leak due to the use of AddRef/Release
			COutputCallbacks* pOutputCallbacks = new COutputCallbacks;
			if (pOutputCallbacks != NULL)
			{
				// Apply IDebugOutputCallbacks to the IDebugClient
				if (SUCCEEDED(pDebugCreated->SetOutputMask(pOutputCallbacks->SupportedMask())) && 
					SUCCEEDED(pDebugCreated->SetOutputCallbacks(pOutputCallbacks)))
				{
					// Execute and display 'vertarget'
					pDebugControl->Execute(DEBUG_OUTCTL_THIS_CLIENT | DEBUG_OUTCTL_NOT_LOGGED, "vertarget", DEBUG_EXECUTE_NOT_LOGGED | DEBUG_EXECUTE_NO_REPEAT);
					pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_OTHER_CLIENTS, DEBUG_OUTPUT_NORMAL, "[Normal]\n%s\n", pOutputCallbacks->BufferNormal() ? pOutputCallbacks->BufferNormal() : "");
					pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_OTHER_CLIENTS, DEBUG_OUTPUT_ERROR , "[Error] \n%s\n", pOutputCallbacks->BufferError() ? pOutputCallbacks->BufferError() : "");

					// Clear the callback buffers
					pOutputCallbacks->Clear();

					// Execute and display the passed command
					pDebugControl->Execute(DEBUG_OUTCTL_THIS_CLIENT | DEBUG_OUTCTL_NOT_LOGGED, args, DEBUG_EXECUTE_NOT_LOGGED | DEBUG_EXECUTE_NO_REPEAT);
					pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_OTHER_CLIENTS, DEBUG_OUTPUT_NORMAL, "[Normal]\n%s\n", pOutputCallbacks->BufferNormal() ? pOutputCallbacks->BufferNormal() : "");
					pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_OTHER_CLIENTS, DEBUG_OUTPUT_ERROR , "[Error] \n%s\n", pOutputCallbacks->BufferError() ? pOutputCallbacks->BufferError() : "");

					pDebugCreated->SetOutputCallbacks(NULL); // It is a best practice to release callbacks before the client release
				}
				pOutputCallbacks->Release();
			}
			pDebugControl->Release();
		}
		pDebugCreated->Release();
	}
	return S_OK;
}

HRESULT CALLBACK 
callbacktoggle(PDEBUG_CLIENT pDebugClient, PCSTR args)
{
	IDebugControl* pDebugControl;
	if (SUCCEEDED(pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
	{
		// Remember the original mask and callback
		ULONG ulOriginalMask;
		IDebugOutputCallbacks* pOriginalCallback;
		if (SUCCEEDED(pDebugClient->GetOutputMask(&ulOriginalMask)) &&
			SUCCEEDED(pDebugClient->GetOutputCallbacks(&pOriginalCallback)))
		{
			// Create our IDebugOutputCallbacks based object
			#pragma warning( disable : 28197 ) // PREfast sees the 'new' as a leak due to the use of AddRef/Release
			COutputCallbacks* pOutputCallbacks = new COutputCallbacks;
			if (pOutputCallbacks != NULL)
			{
				// Apply IDebugOutputCallbacks to the IDebugClient
				if (SUCCEEDED(pDebugClient->SetOutputMask(pOutputCallbacks->SupportedMask())) && 
					SUCCEEDED(pDebugClient->SetOutputCallbacks(pOutputCallbacks)))
				{
					// Execute 'vertarget'
					pDebugControl->Execute(DEBUG_OUTCTL_THIS_CLIENT | DEBUG_OUTCTL_NOT_LOGGED, "vertarget", DEBUG_EXECUTE_NOT_LOGGED | DEBUG_EXECUTE_NO_REPEAT);

					// Revert the mask and callback so we can do some output
					pDebugClient->SetOutputMask(ulOriginalMask);
					pDebugClient->SetOutputCallbacks(pOriginalCallback);

					// Display 'vertarget'
					pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_CLIENTS, DEBUG_OUTPUT_NORMAL, "[Normal]\n%s\n", pOutputCallbacks->BufferNormal() ? pOutputCallbacks->BufferNormal() : "");
					pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_CLIENTS, DEBUG_OUTPUT_ERROR , "[Error] \n%s\n", pOutputCallbacks->BufferError() ? pOutputCallbacks->BufferError() : "");

					// Go back to our mask and callback so we can do the private callback
					pDebugClient->SetOutputMask(pOutputCallbacks->SupportedMask());
					pDebugClient->SetOutputCallbacks(pOutputCallbacks);

					// Clear the callback buffers
					pOutputCallbacks->Clear();

					// Execute the passed command
					pDebugControl->Execute(DEBUG_OUTCTL_THIS_CLIENT | DEBUG_OUTCTL_NOT_LOGGED, args, DEBUG_EXECUTE_NOT_LOGGED | DEBUG_EXECUTE_NO_REPEAT);

					// Revert the mask and callback so we can do some output
					pDebugClient->SetOutputMask(ulOriginalMask);
					pDebugClient->SetOutputCallbacks(pOriginalCallback);

					// Display he passed command
					pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_CLIENTS, DEBUG_OUTPUT_NORMAL, "[Normal]\n%s\n", pOutputCallbacks->BufferNormal() ? pOutputCallbacks->BufferNormal() : "");
					pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_CLIENTS, DEBUG_OUTPUT_ERROR , "[Error] \n%s\n", pOutputCallbacks->BufferError() ? pOutputCallbacks->BufferError() : "");
				}

				// Revert the mask (again) for the case where the mask was set but the callback wasn't
				pDebugClient->SetOutputMask(ulOriginalMask);

				pOutputCallbacks->Release();
			}
		}
		pDebugControl->Release();
	}

	return S_OK;
}
