// MyExt.cpp

#include <stdio.h>
#include "dbgexts.h"
#include "outputcallbacks.h"

/***************************************************************************************************************************************************************/
/*************************************************************** O U T C T L   E X A M P L E S *****************************************************************/
/***************************************************************************************************************************************************************/

HRESULT CALLBACK 
outctlpassed(PDEBUG_CLIENT pDebugClient, PCSTR args)
{
	UNREFERENCED_PARAMETER(args);

	IDebugControl* pDebugControl;
	if (SUCCEEDED(pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
	{
		pDebugControl->ControlledOutput(DEBUG_OUTCTL_THIS_CLIENT,       DEBUG_OUTPUT_NORMAL, "DEBUG_OUTCTL_THIS_CLIENT\n");
		pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_CLIENTS,       DEBUG_OUTPUT_NORMAL, "DEBUG_OUTCTL_ALL_CLIENTS\n");
		pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_OTHER_CLIENTS, DEBUG_OUTPUT_NORMAL, "DEBUG_OUTCTL_ALL_OTHER_CLIENTS\n");
		pDebugControl->ControlledOutput(DEBUG_OUTCTL_IGNORE,            DEBUG_OUTPUT_NORMAL, "DEBUG_OUTCTL_IGNORE\n");
		pDebugControl->ControlledOutput(DEBUG_OUTCTL_LOG_ONLY,          DEBUG_OUTPUT_NORMAL, "DEBUG_OUTCTL_LOG_ONLY\n");

		pDebugControl->Release();
	}
	return S_OK;
}

HRESULT CALLBACK 
outctlcreate(PDEBUG_CLIENT pDebugClient, PCSTR args)
{
	UNREFERENCED_PARAMETER(args);

	IDebugClient* pDebugCreated;
	if (SUCCEEDED(pDebugClient->CreateClient(&pDebugCreated)))
	{
		IDebugControl* pDebugControl;
		if (SUCCEEDED(pDebugCreated->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
		{
			pDebugControl->ControlledOutput(DEBUG_OUTCTL_THIS_CLIENT,       DEBUG_OUTPUT_NORMAL, "DEBUG_OUTCTL_THIS_CLIENT\n");
			pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_CLIENTS,       DEBUG_OUTPUT_NORMAL, "DEBUG_OUTCTL_ALL_CLIENTS\n");
			pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_OTHER_CLIENTS, DEBUG_OUTPUT_NORMAL, "DEBUG_OUTCTL_ALL_OTHER_CLIENTS\n");
			pDebugControl->ControlledOutput(DEBUG_OUTCTL_IGNORE,            DEBUG_OUTPUT_NORMAL, "DEBUG_OUTCTL_IGNORE\n");
			pDebugControl->ControlledOutput(DEBUG_OUTCTL_LOG_ONLY,          DEBUG_OUTPUT_NORMAL, "DEBUG_OUTCTL_LOG_ONLY\n");

			pDebugControl->Release();
		}
		pDebugCreated->Release();
	}
	return S_OK;
}

HRESULT CALLBACK 
outctlthis(PDEBUG_CLIENT pDebugClient, PCSTR args)
{
	UNREFERENCED_PARAMETER(args);

	IDebugClient* pDebugCreated;
	if (SUCCEEDED(pDebugClient->CreateClient(&pDebugCreated)))
	{
		IDebugControl* pDebugControl;
		if (SUCCEEDED(pDebugCreated->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
		{
			pDebugControl->ControlledOutput(DEBUG_OUTCTL_THIS_CLIENT | DEBUG_OUTCTL_NOT_LOGGED, DEBUG_OUTPUT_NORMAL, "DEBUG_OUTCTL_THIS_CLIENT | DEBUG_OUTCTL_NOT_LOGGED\n");
			pDebugControl->Release();
		}
		pDebugCreated->Release();
	}
	return S_OK;
}

HRESULT CALLBACK 
outctlexecute(PDEBUG_CLIENT pDebugClient, PCSTR args)
{
	IDebugClient* pDebugCreated;
	if (SUCCEEDED(pDebugClient->CreateClient(&pDebugCreated)))
	{
		IDebugControl* pDebugControl;
		if (SUCCEEDED(pDebugCreated->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
		{
			pDebugControl->Execute(DEBUG_OUTCTL_THIS_CLIENT | DEBUG_OUTCTL_NOT_LOGGED, args, DEBUG_EXECUTE_NOT_LOGGED | DEBUG_EXECUTE_NO_REPEAT);
			pDebugControl->Release();
		}
		pDebugCreated->Release();
	}
	return S_OK;
}

/***************************************************************************************************************************************************************/
/*********************************************************** C A L L B A C K   E X A M P L E S *****************************************************************/
/***************************************************************************************************************************************************************/

HRESULT CALLBACK 
callbackecho(PDEBUG_CLIENT pDebugClient, PCSTR args)
{
	IDebugClient* pDebugCreated;
	if (SUCCEEDED(pDebugClient->CreateClient(&pDebugCreated)))
	{
		IDebugControl* pDebugControl;
		if (SUCCEEDED(pDebugCreated->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
		{
			// Create our IDebugOutputCallbacks based object
			#pragma warning( disable : 28197 ) // PREfast sees the 'new' as a leak due to the use of AddRef/Release
			COutputCallbacks* pOutputCallbacks = new COutputCallbacks;
			if (pOutputCallbacks != NULL)
			{
				// Apply IDebugOutputCallbacks to the IDebugClient
				if (SUCCEEDED(pDebugCreated->SetOutputMask(pOutputCallbacks->SupportedMask())) && 
					SUCCEEDED(pDebugCreated->SetOutputCallbacks(pOutputCallbacks)))
				{
					// Execute and display 'vertarget'
					pDebugControl->Execute(DEBUG_OUTCTL_THIS_CLIENT | DEBUG_OUTCTL_NOT_LOGGED, "vertarget", DEBUG_EXECUTE_NOT_LOGGED | DEBUG_EXECUTE_NO_REPEAT);
					pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_OTHER_CLIENTS, DEBUG_OUTPUT_NORMAL, "[Normal]\n%s\n", pOutputCallbacks->BufferNormal() ? pOutputCallbacks->BufferNormal() : "");
					pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_OTHER_CLIENTS, DEBUG_OUTPUT_ERROR , "[Error] \n%s\n", pOutputCallbacks->BufferError() ? pOutputCallbacks->BufferError() : "");

					// Clear the callback buffers
					pOutputCallbacks->Clear();

					// Execute and display the passed command
					pDebugControl->Execute(DEBUG_OUTCTL_THIS_CLIENT | DEBUG_OUTCTL_NOT_LOGGED, args, DEBUG_EXECUTE_NOT_LOGGED | DEBUG_EXECUTE_NO_REPEAT);
					pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_OTHER_CLIENTS, DEBUG_OUTPUT_NORMAL, "[Normal]\n%s\n", pOutputCallbacks->BufferNormal() ? pOutputCallbacks->BufferNormal() : "");
					pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_OTHER_CLIENTS, DEBUG_OUTPUT_ERROR , "[Error] \n%s\n", pOutputCallbacks->BufferError() ? pOutputCallbacks->BufferError() : "");

					pDebugCreated->SetOutputCallbacks(NULL); // It is a best practice to release callbacks before the client release
				}
				pOutputCallbacks->Release();
			}
			pDebugControl->Release();
		}
		pDebugCreated->Release();
	}
	return S_OK;
}

HRESULT CALLBACK 
callbacktoggle(PDEBUG_CLIENT pDebugClient, PCSTR args)
{
	IDebugControl* pDebugControl;
	if (SUCCEEDED(pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
	{
		// Remember the original mask and callback
		ULONG ulOriginalMask;
		IDebugOutputCallbacks* pOriginalCallback;
		if (SUCCEEDED(pDebugClient->GetOutputMask(&ulOriginalMask)) &&
			SUCCEEDED(pDebugClient->GetOutputCallbacks(&pOriginalCallback)))
		{
			// Create our IDebugOutputCallbacks based object
			#pragma warning( disable : 28197 ) // PREfast sees the 'new' as a leak due to the use of AddRef/Release
			COutputCallbacks* pOutputCallbacks = new COutputCallbacks;
			if (pOutputCallbacks != NULL)
			{
				// Apply IDebugOutputCallbacks to the IDebugClient
				if (SUCCEEDED(pDebugClient->SetOutputMask(pOutputCallbacks->SupportedMask())) && 
					SUCCEEDED(pDebugClient->SetOutputCallbacks(pOutputCallbacks)))
				{
					// Execute 'vertarget'
					pDebugControl->Execute(DEBUG_OUTCTL_THIS_CLIENT | DEBUG_OUTCTL_NOT_LOGGED, "vertarget", DEBUG_EXECUTE_NOT_LOGGED | DEBUG_EXECUTE_NO_REPEAT);

					// Revert the mask and callback so we can do some output
					pDebugClient->SetOutputMask(ulOriginalMask);
					pDebugClient->SetOutputCallbacks(pOriginalCallback);

					// Display 'vertarget'
					pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_CLIENTS, DEBUG_OUTPUT_NORMAL, "[Normal]\n%s\n", pOutputCallbacks->BufferNormal() ? pOutputCallbacks->BufferNormal() : "");
					pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_CLIENTS, DEBUG_OUTPUT_ERROR , "[Error] \n%s\n", pOutputCallbacks->BufferError() ? pOutputCallbacks->BufferError() : "");

					// Go back to our mask and callback so we can do the private callback
					pDebugClient->SetOutputMask(pOutputCallbacks->SupportedMask());
					pDebugClient->SetOutputCallbacks(pOutputCallbacks);

					// Clear the callback buffers
					pOutputCallbacks->Clear();

					// Execute the passed command
					pDebugControl->Execute(DEBUG_OUTCTL_THIS_CLIENT | DEBUG_OUTCTL_NOT_LOGGED, args, DEBUG_EXECUTE_NOT_LOGGED | DEBUG_EXECUTE_NO_REPEAT);

					// Revert the mask and callback so we can do some output
					pDebugClient->SetOutputMask(ulOriginalMask);
					pDebugClient->SetOutputCallbacks(pOriginalCallback);

					// Display he passed command
					pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_CLIENTS, DEBUG_OUTPUT_NORMAL, "[Normal]\n%s\n", pOutputCallbacks->BufferNormal() ? pOutputCallbacks->BufferNormal() : "");
					pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_CLIENTS, DEBUG_OUTPUT_ERROR , "[Error] \n%s\n", pOutputCallbacks->BufferError() ? pOutputCallbacks->BufferError() : "");
				}

				// Revert the mask (again) for the case where the mask was set but the callback wasn't
				pDebugClient->SetOutputMask(ulOriginalMask);

				pOutputCallbacks->Release();
			}
		}
		pDebugControl->Release();
	}

	return S_OK;
}

/***************************************************************************************************************************************************************/
/************************************************************ P A R S I N G    E X A M P L E S *****************************************************************/
/***************************************************************************************************************************************************************/
BOOL PreferDML(IDebugClient* pDebugClient)
{
	BOOL bPreferDML = FALSE;
	IDebugControl* pDebugControl;
	if (SUCCEEDED(pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)& pDebugControl)))
	{
		ULONG ulOptions = 0;
		if (SUCCEEDED(pDebugControl->GetEngineOptions(&ulOptions)))
		{
			bPreferDML = (ulOptions & DEBUG_ENGOPT_PREFER_DML);
		}
		pDebugControl->Release();
	}
	return bPreferDML;
}

BOOL AbilityDML(IDebugClient* pDebugClient)
{
	BOOL bAbilityDML = FALSE;
	IDebugAdvanced2* pDebugAdvanced2;
	if (SUCCEEDED(pDebugClient->QueryInterface(__uuidof(IDebugAdvanced2), (void **)& pDebugAdvanced2)))
	{
		HRESULT hr = 0;
		if (SUCCEEDED(hr = pDebugAdvanced2->Request(DEBUG_REQUEST_CURRENT_OUTPUT_CALLBACKS_ARE_DML_AWARE, NULL, 0, NULL, 0, NULL)))
		{
			if (hr == S_OK) bAbilityDML = TRUE;
		}
		pDebugAdvanced2->Release();
	}
	return bAbilityDML;
}

/*
0:000> .extmatch /e *\sos.dll dumpstackobjects
!C:\Windows\Microsoft.NET\Framework\v2.0.50727\sos.dll.DumpStackObjects
!C:\Windows\Microsoft.NET\Framework\v2.0.50727\sos.dll.dumpstackobjects

0:000> .extmatch /e *\sos.dll dumpstackobjects
!C:\Windows\Microsoft.NET\Framework64\v2.0.50727\sos.dll.DumpStackObjects
!C:\Windows\Microsoft.NET\Framework64\v2.0.50727\sos.dll.dumpstackobjects
*/
HRESULT LoadSOS(IDebugControl* pDebugControl, COutputCallbacks* pOutputCallbacks, PCSTR szVersion, PCSTR szCommand, PSTR szResult, size_t nResult)
{
	HRESULT hr = E_FAIL;

	BOOL bFound = false;
	BOOL bLoadbyTried = false;

	char szExtMatch[1024];
	char szVersioned[1024];
	char szNoVersion[1024];
	char* pos = NULL;

	if ((pDebugControl == NULL) || (pOutputCallbacks == NULL) || (szVersion == NULL) || (szVersion[0] == '\0') || (szCommand == NULL) || (szCommand[0] == '\0'))
		return E_INVALIDARG;

	if ((sprintf_s(szExtMatch, 1024, ".extmatch /e *\\sos.dll %s", szCommand) < 0) || 
		(sprintf_s(szVersioned, 1024, "\\%s\\sos.dll.%s", szVersion, szCommand) < 0) ||
		(sprintf_s(szNoVersion, 1024, "\\sos.dll.%s", szCommand) < 0))
		return E_OUTOFMEMORY;

LoadbyRetry:
	pOutputCallbacks->Clear();
	if ((SUCCEEDED(pDebugControl->Execute(DEBUG_OUTCTL_THIS_CLIENT | DEBUG_OUTCTL_NOT_LOGGED, szExtMatch, DEBUG_EXECUTE_NOT_LOGGED | DEBUG_EXECUTE_NO_REPEAT))) &&
		(pOutputCallbacks->BufferError() == NULL) && (pOutputCallbacks->BufferNormal() != NULL))
	{	// We have some type of match
		if ((pos = strstr(pOutputCallbacks->BufferNormal(), szVersioned)) != NULL)
		{	// Found versioned match
			bFound = true;
			hr = S_OK;
		}
		else if ((pos = strstr(pOutputCallbacks->BufferNormal(), szNoVersion)) != NULL)
		{	// Found non-versioned match
			bFound = true;
			hr = S_FALSE;
		}
	}

	if (!bFound && !bLoadbyTried)
	{
		bLoadbyTried = true;
		pOutputCallbacks->Clear();
		if ((SUCCEEDED(pDebugControl->Execute(DEBUG_OUTCTL_THIS_CLIENT | DEBUG_OUTCTL_NOT_LOGGED, ".loadby sos.dll mscorwks", DEBUG_EXECUTE_NOT_LOGGED | DEBUG_EXECUTE_NO_REPEAT))) &&
			(pOutputCallbacks->BufferError() == NULL))
		{	// Loadby succeeded
			goto LoadbyRetry;
		}
	}

	if (bFound)
	{
		char* szStart = pOutputCallbacks->BufferNormal();
		char* szEnd = strchr(pos, 0x0A);
		char* szTemp = NULL;

		if (szEnd != NULL)
			*szEnd = '\0'; // Note, this modifies the buffer

		szTemp = strrchr(szStart, 0x0A);
		if (szTemp != NULL)
			szStart = szTemp + 1;

		if (strcpy_s(szResult, nResult, szStart) != 0)
			hr = E_OUTOFMEMORY;
	}

	return hr;
}

HRESULT ParseDSO(IDebugControl* pDebugControl, COutputCallbacks* pOutputCallbacks, PCSTR szNormal, char** szReturn)
{
	HRESULT hr = E_FAIL;

	if ((pDebugControl == NULL) || (pOutputCallbacks == NULL) || (szNormal == NULL) || (szReturn == NULL))
		return E_INVALIDARG;

	*szReturn = NULL;

	char szCommand[1024];
	if (SUCCEEDED(LoadSOS(pDebugControl, pOutputCallbacks, "v2.0.50727", "dumpobj", szCommand, 1024)))
	{
		// Convert 1234567890123456 to "<link cmd="<dumpobject command> 1234567890123456">1234567890123456</link>"
		size_t nNormal = strlen(szNormal);
		size_t nResult = nNormal + (nNormal / 18 /*approx lines*/ * (strlen(szCommand) + 40 /*addition text*/)) + 0x1000 /*safety*/;
		size_t nPos = 0;
		char* szResult = (char*)malloc(nResult);
		
		const char* szStart = NULL;
		if ((szResult != NULL) &&
			((szStart=strstr(szNormal, "Object           Name")) != NULL) &&	// The x64 header...
			((szStart=strchr(szStart, 0x0A)) != NULL))			// 'End of the line' of the header...
		{	// x64
			szStart++;
			memcpy_s(&szResult[nPos], nResult-nPos, szNormal, szStart-szNormal);
			nPos += szStart-szNormal;

			while (szStart < szNormal+nNormal)
			{
				const char* szEnd = strchr(szStart, 0x0A);		// End of the line
				if (szEnd == NULL)								// End of the string
					szEnd = szNormal+nNormal;

				if (((szStart+33) < szEnd) && (szStart[16] == ' '))
				{	// x64
					memcpy_s(&szResult[nPos], nResult-nPos, szStart, 17); nPos += 17;
					memcpy_s(&szResult[nPos], nResult-nPos, "<link cmd=\"", 11); nPos += 11;
					memcpy_s(&szResult[nPos], nResult-nPos, szCommand, strlen(szCommand)); nPos += strlen(szCommand);
					memcpy_s(&szResult[nPos], nResult-nPos, " ", 1); nPos += 1;
					memcpy_s(&szResult[nPos], nResult-nPos, &szStart[17], 16); nPos += 16;
					memcpy_s(&szResult[nPos], nResult-nPos, "\">", 2); nPos += 2;
					memcpy_s(&szResult[nPos], nResult-nPos, &szStart[17], 16); nPos += 16;
					memcpy_s(&szResult[nPos], nResult-nPos, "</link>", 7); nPos += 7;
					memcpy_s(&szResult[nPos], nResult-nPos, &szStart[33], szEnd-szStart-32); nPos += szEnd-szStart-32; // TODO: Escape Sequence
				}
				else
				{
					memcpy_s(&szResult[nPos], nResult-nPos, szStart, szEnd-szStart+1); nPos += szEnd-szStart+1; // TODO: Escape Sequence
				}
				szStart = szEnd + 1;
			}
			szResult[nPos] = '\0';
			*szReturn = szResult;
			hr = S_OK;
		}
		else if ((szResult != NULL) &&
				((szStart=strstr(szNormal, "Object   Name")) != NULL) &&	// The x86 header...
				((szStart=strchr(szStart, 0x0A)) != NULL))			// 'End of the line' of the header...
		{	// x86
			szStart++;
			memcpy_s(&szResult[nPos], nResult-nPos, szNormal, szStart-szNormal);
			nPos += szStart-szNormal;

			while (szStart < szNormal+nNormal)
			{
				const char* szEnd = strchr(szStart, 0x0A);		// End of the line
				if (szEnd == NULL)								// End of the string
					szEnd = szNormal+nNormal;

				if (((szStart+17) < szEnd) && (szStart[8] == ' '))
				{
					memcpy_s(&szResult[nPos], nResult-nPos, szStart, 9); nPos += 9;
					memcpy_s(&szResult[nPos], nResult-nPos, "<link cmd=\"", 11); nPos += 11;
					memcpy_s(&szResult[nPos], nResult-nPos, szCommand, strlen(szCommand)); nPos += strlen(szCommand);
					memcpy_s(&szResult[nPos], nResult-nPos, " ", 1); nPos += 1;
					memcpy_s(&szResult[nPos], nResult-nPos, &szStart[9], 8); nPos += 8;
					memcpy_s(&szResult[nPos], nResult-nPos, "\">", 2); nPos += 2;
					memcpy_s(&szResult[nPos], nResult-nPos, &szStart[9], 8); nPos += 8;
					memcpy_s(&szResult[nPos], nResult-nPos, "</link>", 7); nPos += 7;
					memcpy_s(&szResult[nPos], nResult-nPos, &szStart[17], szEnd-szStart-16); nPos += szEnd-szStart-16; // TODO: Escape Sequence
				}
				else
				{
					memcpy_s(&szResult[nPos], nResult-nPos, szStart, szEnd-szStart+1); nPos += szEnd-szStart+1; // TODO: Escape Sequence
				}
				szStart = szEnd + 1;
			}
			szResult[nPos] = '\0';
			*szReturn = szResult;
			hr = S_OK;
		}

		if ((*szReturn == NULL) && (szResult != NULL))
			free(szResult);
	}
	return hr;
}

HRESULT CALLBACK 
dsox(PDEBUG_CLIENT pDebugClient, PCSTR args)
{
	UNREFERENCED_PARAMETER(args);

	IDebugClient* pDebugCreated;
	if (SUCCEEDED(pDebugClient->CreateClient(&pDebugCreated)))
	{
		IDebugControl* pDebugControl;
		if (SUCCEEDED(pDebugCreated->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
		{
			// Create our IDebugOutputCallbacks based object
			#pragma warning( disable : 28197 ) // PREfast sees the 'new' as a leak due to the use of AddRef/Release
			COutputCallbacks* pOutputCallbacks = new COutputCallbacks;
			if (pOutputCallbacks != NULL)
			{
				// Apply IDebugOutputCallbacks to the IDebugClient
				if (SUCCEEDED(pDebugCreated->SetOutputMask(pOutputCallbacks->SupportedMask())) && 
					SUCCEEDED(pDebugCreated->SetOutputCallbacks(pOutputCallbacks)))
				{
					char szCommand[1024];
					if (SUCCEEDED(LoadSOS(pDebugControl, pOutputCallbacks, "v2.0.50727", "dumpstackobjects", szCommand, 1024)))
					{
						// Execute and display 'dumpstackobjects'
						pOutputCallbacks->Clear();
						if ((SUCCEEDED(pDebugControl->Execute(DEBUG_OUTCTL_THIS_CLIENT | DEBUG_OUTCTL_NOT_LOGGED, szCommand, DEBUG_EXECUTE_NOT_LOGGED | DEBUG_EXECUTE_NO_REPEAT))) &&
							(pOutputCallbacks->BufferError() == NULL) && (pOutputCallbacks->BufferNormal() != NULL))
						{
							// Output DML?  -- I'm to ignoring the .prefer_dml preference since the point of this command is to add DML to !dumpstackobjects
							if (/*PreferDML(pDebugClient) &&*/ AbilityDML(pDebugClient))
							{
								// Save the normal output so the callback can be reused
								size_t nNormal = strlen(pOutputCallbacks->BufferNormal())+1;
								char* szNormal = (char*)malloc(nNormal);
								if (szNormal != NULL)
								{
									memcpy_s(szNormal, nNormal, pOutputCallbacks->BufferNormal(), nNormal);
									pOutputCallbacks->Clear();
									char* szDML = NULL;
									if (SUCCEEDED(ParseDSO(pDebugControl, pOutputCallbacks, szNormal, &szDML)))
									{	// Output the DML
										pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_OTHER_CLIENTS | DEBUG_OUTCTL_DML, DEBUG_OUTPUT_NORMAL, "%s\n", szDML);
										free(szDML);
									}
									free(szNormal);
								}
							}
							else
							{
								pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_OTHER_CLIENTS, DEBUG_OUTPUT_NORMAL, "%s\n", pOutputCallbacks->BufferNormal());
							}
							pOutputCallbacks->Clear();
						}
					}

					// If the callback has content, something went wrong
					if (pOutputCallbacks->BufferNormal())
						pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_OTHER_CLIENTS, DEBUG_OUTPUT_NORMAL, "%s\n", pOutputCallbacks->BufferNormal());

					if (pOutputCallbacks->BufferError())
						pDebugControl->ControlledOutput(DEBUG_OUTCTL_ALL_OTHER_CLIENTS, DEBUG_OUTPUT_ERROR, "%s\n", pOutputCallbacks->BufferError());

					pDebugCreated->SetOutputCallbacks(NULL); // It is a best practice to release callbacks before the client release
				}
				pOutputCallbacks->Release();
			}
			pDebugControl->Release();
		}
		pDebugCreated->Release();
	}
	return S_OK;
}
